#include <list>
#include <array>
#include <vector>
#include <fstream>
#include <iostream>
#include <GL/glew.h>
#include <GL/glut.h>
#include <stdexcept>
#include <glm/glm.hpp>
#include <glm/ext.hpp>

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

#define unused(v) (void)v

//------------------------------------------------------------------------------

class File {
public:
    explicit File(const std::string &filename)
        : m_file(filename.c_str(), std::ios::binary)
    {
        if (!m_file.is_open())
            throw std::runtime_error("Can't open file " + filename);
        m_file.seekg(0, std::ios::end);
        m_size = m_file.tellg();
        m_file.seekg(0, std::ios::beg);
    }

    ~File() {
        close();
    }

    void toBegin() {
        m_file.seekg(0, std::ios::beg);
    }

    bool isEof() const {
        return m_file.eof();
    }

    uint64_t size() const {
        return m_size;
    }

    void close() {
        m_file.close();
    }

    void read(void *data, uint64_t size) {
        m_file.read(reinterpret_cast<char*>(data), size);
    }

    std::vector<uint8_t> readAllBytes() {
        toBegin();
        std::vector<uint8_t> d(size());
        read(d.data(), size());
        return std::move(d);
    }

    std::string readAll() {
        toBegin();
        std::vector<char> str(size() + 1, 0);
        read(str.data(), size());
        std::string s(str.data());
        return std::move(s);
    }

    std::string readLine() {
        std::string str;
        while (!isEof()) {
            char c;
            read(&c, 1);
            if (c == '\r')
                continue;
            if (c == 0 || c == '\n')
                break;
            str += c;
        }
        return str;
    }

private:
    std::ifstream m_file;
    uint64_t m_size;
};

//------------------------------------------------------------------------------

glm::vec4 viewport() {
    GLint vp[4];
    glGetIntegerv(GL_VIEWPORT, vp);
    return glm::vec4(vp[0], vp[1], vp[2], vp[3]);
}

//------------------------------------------------------------------------------

struct Vertex {
    Vertex()
    {}

    Vertex(const glm::vec3 &p, const glm::vec3 &n, const glm::vec2 &t)
        : position(p), normal(n), texcoord(t)
    {}

    glm::vec3 position;
    glm::vec3 normal;
    glm::vec2 texcoord;
    glm::vec3 tangent;
    glm::vec3 bitangent;
};
typedef std::vector<Vertex> VertexList;

//------------------------------------------------------------------------------

class GlResource {
public:
    virtual ~GlResource() {}

    GLuint glHandle() const {
        return m_glHandle;
    }
protected:
    GLuint m_glHandle;
};

//------------------------------------------------------------------------------

class Texture : public GlResource {
protected:
    enum Filtration {
        NearestFiltration,
        LinearFiltration,
        MipMapFiltration
    };
    enum Wrapping {
        ClampWrapping,
        RepeatClamping
    };
public:
    Texture(GLenum target = GL_TEXTURE_2D)
        : m_target(target)
    {
        glGenTextures(1, &m_glHandle);
    }

    ~Texture() {
        glDeleteTextures(1, &m_glHandle);
    }

    void bind(uint8_t layer = 0) {
        glActiveTexture(GL_TEXTURE0 + layer);
        glBindTexture(m_target, m_glHandle);
    }
protected:
    void setParameters(Filtration filtration, Wrapping wrapping) {
        GLenum f;
        GLenum w;
        switch (filtration) {
        case NearestFiltration:
            f = GL_NEAREST;
            break;
        case LinearFiltration:
            f = GL_LINEAR;
            break;
        case MipMapFiltration:
            f = GL_LINEAR_MIPMAP_LINEAR;
            break;
        }
        switch (wrapping) {
        case ClampWrapping:
            w = GL_CLAMP_TO_EDGE;
            break;
        case RepeatClamping:
            w = GL_REPEAT;
            break;
        }
        glTexParameteri(m_target, GL_TEXTURE_MAG_FILTER, f);
        glTexParameteri(m_target, GL_TEXTURE_MIN_FILTER, f);
        glTexParameteri(m_target, GL_TEXTURE_WRAP_S, w);
        glTexParameteri(m_target, GL_TEXTURE_WRAP_T, w);
        glTexParameteri(m_target, GL_TEXTURE_WRAP_R, w);
    }
protected:
    GLenum m_target;
};

//------------------------------------------------------------------------------

class ColorAttachmentTexture : public Texture {
public:
    ColorAttachmentTexture() {
        bind();
        auto vp = viewport();
        glTexImage2D(m_target, 0, GL_RGBA16F, vp[2], vp[3], 0, GL_RGBA, GL_FLOAT, 0);
        setParameters(NearestFiltration, ClampWrapping);
    }
};

//------------------------------------------------------------------------------

class DepthAttachmentTexture : public Texture {
public:
    DepthAttachmentTexture() {
        bind();
        auto vp = viewport();
        glTexImage2D(m_target, 0, GL_DEPTH_COMPONENT32F, vp[2], vp[3], 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
        setParameters(NearestFiltration, ClampWrapping);
    }
};

//------------------------------------------------------------------------------

class ColorTexture : public Texture {
public:
    ColorTexture(uint8_t r, uint8_t g, uint8_t b, uint8_t a) {
        int color = a;
        color = color << 8;
        color |= b;
        color = color << 8;
        color |= g;
        color = color << 8;
        color |= r;

        std::vector<int> texBuf(1024, color);
        bind();
        glTexImage2D(m_target, 0, GL_RGBA, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, texBuf.data());
        setParameters(NearestFiltration, ClampWrapping);
        glGenerateMipmap(m_target);
    }

    static ColorTexture *whiteTexture() {
        return new ColorTexture(255, 255, 255, 255);
    }

    static ColorTexture *normalIdentTexture() {
        return new ColorTexture(128, 127, 254, 255);
    }
};

//------------------------------------------------------------------------------

class ImageTexture : public Texture {
public:
    explicit ImageTexture(const std::string &filename) {
        auto buf = File(filename).readAllBytes();
        int w, h, c;
        auto texData = stbi_load_from_memory(buf.data(), buf.size(), &w, &h, &c, 4);
        bind();
        glTexImage2D(m_target, 0, c == 3 ? GL_RGB : GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, texData);
        free(texData);
        setParameters(MipMapFiltration, RepeatClamping);
        glGenerateMipmap(m_target);
    }
};

//------------------------------------------------------------------------------

struct Material {
    Texture *diffuseTexture = 0;
    Texture *normalTexture = 0;
    Texture *lightmapTexture = 0;
};

//------------------------------------------------------------------------------

class Shader : public GlResource{
public:
    Shader(const std::string &source) {
        m_glHandle = glCreateProgram();
        GLuint vs = 0;
        GLuint fs = 0;
        try
        {
            vs = createShader("#define _vertex_\n" + source, GL_VERTEX_SHADER);
            fs = createShader("#define _fragment_\n" + source, GL_FRAGMENT_SHADER);
        }
        catch (...)
        {
            glDeleteShader(vs);
            glDeleteShader(fs);
            throw;
        }
        glAttachShader(m_glHandle, vs);
        glAttachShader(m_glHandle, fs);
        glLinkProgram(m_glHandle);
        glDeleteShader(vs);
        glDeleteShader(fs);
    }

    ~Shader() {
        glDeleteProgram(m_glHandle);
    }

    void use() {
        if (s_shader != this) {
            if (s_shader)
                s_shader->disableAttributes();
            s_shader = this;
            glUseProgram(m_glHandle);
            enableAttributes();
        }
    }

    virtual void setupVertexBuffer(const void *ptr) { unused(ptr); }

    virtual void setupMaterial(const Material &material) { unused(material); }

protected:
    virtual void enableAttributes() {}

    virtual void disableAttributes() {}

    GLuint createShader(std::string source, GLenum type)
    {
        source =
            "#version 300 es            \n"
            "precision mediump float;   \n"
            + source;

        GLuint shader = glCreateShader(type);

        const GLchar *src = source.c_str();
        glShaderSource(shader, 1, &src, NULL);
        glCompileShader(shader);
        GLint isCompiled = 0;
        glGetShaderiv(shader, GL_COMPILE_STATUS, &isCompiled);
        if(isCompiled == GL_FALSE)
        {
            GLint maxLength = 0;
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &maxLength);
            std::vector<GLchar> infoLog(maxLength);
            glGetShaderInfoLog(shader, maxLength, &maxLength, &infoLog[0]);
            glDeleteShader(shader);
            std::string message = infoLog.data();
            message += "\n" + source;
            throw std::runtime_error(message);
        }
        return shader;
    }

    GLint getAttribLocation(const char *name) const {
        return glGetAttribLocation(m_glHandle, name);
    }

    GLint getUniformLocation(const char *name) const {
        return glGetUniformLocation(m_glHandle, name);
    }

    void enableAttribute(GLint attr) {
        // must be used shader
        glEnableVertexAttribArray(attr);
    }

    void disableAttribute(GLint attr) {
        // must be used shader
        glDisableVertexAttribArray(attr);
    }

    void setVertexPointer(GLint attr, uint8_t count, uint16_t stride, const void *data) {
        use();
        glVertexAttribPointer(attr, count, GL_FLOAT, GL_FALSE, stride, data);
    }

    void setUniformMatrix4(GLint unif, const glm::mat4 &matrix) {
        use();
        glUniformMatrix4fv(unif, 1, GL_FALSE, glm::value_ptr(matrix));
    }

    void setUniformMatrix3(GLint unif, const glm::mat3 &matrix) {
        use();
        glUniformMatrix3fv(unif, 1, GL_FALSE, glm::value_ptr(matrix));
    }

    void setUniformFloat4(GLint unif, const glm::vec4 &value) {
        use();
        glUniform4fv(unif, 1, glm::value_ptr(value));
    }

    void setUniformFloat3(GLint unif, const glm::vec3 &value) {
        use();
        glUniform3fv(unif, 1, glm::value_ptr(value));
    }

    void setTexture(GLint unif, Texture *texture, uint8_t layer) {
        use();
        if (!texture)
            return;
        texture->bind(layer);
        glUniform1i(unif, layer);
    }

private:
    static Shader *s_shader;
};

Shader *Shader::s_shader = 0;

//------------------------------------------------------------------------------

class GeometryBufferShader : public Shader {
public:
    GeometryBufferShader()
        : Shader(File("geometry_buffer.glsl").readAll())
    {
        m_attrPosition = getAttribLocation("_attr_position_");
        m_attrNormal = getAttribLocation("_attr_normal_");
        m_attrTexcoord = getAttribLocation("_attr_texcoord_");
        m_attrTangent = getAttribLocation("_attr_tangent_");
        m_attrBitangent = getAttribLocation("_attr_bitangent_");

        m_unifModelMatrix = getUniformLocation("_model_matrix_");
        m_unifNormalMatrix = getUniformLocation("_normal_matrix_");
        m_unifProjViewModelMatrix = getUniformLocation("_proj_view_model_matrix_");
        m_unifDiffuseTexture = getUniformLocation("_diffuse_texture_");
        m_unifNormalTexture = getUniformLocation("_normal_texture_");
        m_unifLightmapTexture = getUniformLocation("_lightmap_texture_");
    }

    // Shader interface
public:
    void setupVertexBuffer(const void *ptr) {
        const uint8_t *d = reinterpret_cast<const uint8_t*>(ptr);
        setVertexPointer(m_attrPosition, 3, sizeof(Vertex), d + 0);
        setVertexPointer(m_attrNormal, 3, sizeof(Vertex), d + 12);
        setVertexPointer(m_attrTexcoord, 2, sizeof(Vertex), d + 24);
        setVertexPointer(m_attrTangent, 3, sizeof(Vertex), d + 32);
        setVertexPointer(m_attrBitangent, 3, sizeof(Vertex), d + 44);
    }

    void setupMatrices(const glm::mat4 &p, const glm::mat4 &v, const glm::mat4 &m) {
        setUniformMatrix4(m_unifProjViewModelMatrix, p * v * m);
        setUniformMatrix4(m_unifModelMatrix, m);
        setUniformMatrix3(m_unifNormalMatrix, glm::mat3(glm::inverseTranspose(m)));
    }

    void setupMaterial(const Material &material) {
        setTexture(m_unifDiffuseTexture, material.diffuseTexture, 0);
        setTexture(m_unifNormalTexture, material.normalTexture, 1);
        setTexture(m_unifLightmapTexture, material.lightmapTexture, 2);
    }
protected:
    void enableAttributes() {
        enableAttribute(m_attrPosition);
        enableAttribute(m_attrNormal);
        enableAttribute(m_attrTexcoord);
        enableAttribute(m_attrTangent);
        enableAttribute(m_attrBitangent);
    }

    void disableAttributes() {
        disableAttribute(m_attrPosition);
        disableAttribute(m_attrNormal);
        disableAttribute(m_attrTexcoord);
        disableAttribute(m_attrTangent);
        disableAttribute(m_attrBitangent);
    }
private:
    GLint m_attrPosition;
    GLint m_attrNormal;
    GLint m_attrTexcoord;
    GLint m_attrTangent;
    GLint m_attrBitangent;
    GLint m_unifProjViewModelMatrix;
    GLint m_unifModelMatrix;
    GLint m_unifNormalMatrix;
    GLint m_unifDiffuseTexture;
    GLint m_unifNormalTexture;
    GLint m_unifLightmapTexture;
};

//------------------------------------------------------------------------------

class LightBufferShader : public Shader {
public:
    LightBufferShader()
        : Shader(File("light_buffer.glsl").readAll())
    {
        m_attrPosition = getAttribLocation("_attr_position_");
        m_attrTexcoord = getAttribLocation("_attr_texcoord_");

        m_unifMaterialDepthTexture = getUniformLocation("_material_depth_texture_");
        m_unifNormalTexture = getUniformLocation("_normal_texture_");
        m_unifColorTexture = getUniformLocation("_color_texture_");
        m_unifInvProjViewMatrix = getUniformLocation("_inv_proj_view_matrix_");
        m_unifCameraPosition = getUniformLocation("_camera_position_");
        m_unifLightPosition = getUniformLocation("_light_position_");
        m_unifLightAttenuation = getUniformLocation("_light_attenuation_");
    }

    // Shader interface
public:
    void setupVertexBuffer(const void *ptr) {
        const uint8_t *d = reinterpret_cast<const uint8_t*>(ptr);
        setVertexPointer(m_attrPosition, 3, sizeof(Vertex), d + 0);
        setVertexPointer(m_attrTexcoord, 2, sizeof(Vertex), d + 24);
    }

    void setupGeometryBuffer(Texture *materialDepth, Texture *normal, Texture *color) {
        setTexture(m_unifMaterialDepthTexture, materialDepth, 0);
        setTexture(m_unifNormalTexture, normal, 1);
        setTexture(m_unifColorTexture, color, 2);
    }

    void setupProjViewMatrix(const glm::mat4 &projViewMatrix) {
        setUniformMatrix4(m_unifInvProjViewMatrix, glm::inverse(projViewMatrix));
    }

    void setupCameraPosition(const glm::vec3 &position) {
        setUniformFloat3(m_unifCameraPosition, position);
    }

    void setupLightParams(const glm::vec3 &position, const glm::vec3 &attenuation) {
        setUniformFloat3(m_unifLightPosition, position);
        setUniformFloat3(m_unifLightAttenuation, attenuation);
    }

protected:
    void enableAttributes() {
        enableAttribute(m_attrPosition);
        enableAttribute(m_attrTexcoord);
    }

    void disableAttributes() {
        disableAttribute(m_attrPosition);
        disableAttribute(m_attrTexcoord);
    }
private:
    GLint m_attrPosition;
    GLint m_attrTexcoord;
    GLint m_unifInvProjViewMatrix;
    GLint m_unifMaterialDepthTexture;
    GLint m_unifNormalTexture;
    GLint m_unifColorTexture;
    GLint m_unifCameraPosition;
    GLint m_unifLightPosition;
    GLint m_unifLightAttenuation;
};

//------------------------------------------------------------------------------

class Framebuffer : public GlResource {
public:
    Framebuffer()
        : m_depthTexture(new DepthAttachmentTexture)
    {
        glGenFramebuffers(1, &m_glHandle);
        bind();
        glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, m_depthTexture->glHandle(), 0);
        unbind();
    }

    ~Framebuffer() {
        glDeleteFramebuffers(1, &m_glHandle);
    }

    virtual void bind() {
        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, m_glHandle);
    }

    void unbind() {
        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, s_defaultFramebuffer);
        GLenum buffers[] = {GL_COLOR_ATTACHMENT0};
        glDrawBuffers(1, buffers);
    }
protected:
    DepthAttachmentTexture *m_depthTexture;
private:
    static GLuint s_defaultFramebuffer;
};

GLuint Framebuffer::s_defaultFramebuffer = 0;

//------------------------------------------------------------------------------

class GeometryBufferFramebuffer : public Framebuffer {
public:
    enum {
        MaterialDepthTexture,
        NormalTexture,
        ColorTexture
    };
public:
    GeometryBufferFramebuffer() {
        bind();
        for (uint64_t i = 0; i < m_textures.size(); ++i) {
            m_textures[i] = new ColorAttachmentTexture;
            glFramebufferTexture2D(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + i, GL_TEXTURE_2D, m_textures[i]->glHandle(), 0);
        }
        unbind();
    }

    void bind() {
        Framebuffer::bind();
        GLenum buffers[] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2};
        glDrawBuffers(3, &buffers[0]);
    }

    ColorAttachmentTexture *texture(uint8_t attachment) const {
        return m_textures[attachment];
    }

private:
    std::array<ColorAttachmentTexture*, 3> m_textures;
};

//------------------------------------------------------------------------------

struct Surface {
    uint64_t start;
    uint64_t count;
    Material material;
};

typedef std::vector<Surface> SurfaceList;

//------------------------------------------------------------------------------

class Mesh : public GlResource {
public:
    Mesh(const VertexList &vertices, const SurfaceList &surfaces)
        : m_vertices(vertices), m_surfaces(surfaces)
    {
        for (uint i = 0; i < m_vertices.size() / 3; ++i)
            calculateTangent(m_vertices[i * 3 + 0], m_vertices[i * 3 + 1], m_vertices[i * 3 + 2]);

        glGenBuffers(1, &m_glHandle);
        bind();
        glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * m_vertices.size(), m_vertices.data(), GL_STATIC_DRAW);
    }

    ~Mesh() {
        glDeleteBuffers(1, &m_glHandle);
    }

    void render(Shader *shader) {
        bind();
        shader->setupVertexBuffer(0);
        for (auto s : m_surfaces) {
            shader->setupMaterial(s.material);
            glDrawArrays(GL_TRIANGLES, s.start, s.count);
        }
    }
private:
    void bind() {
        glBindBuffer(GL_ARRAY_BUFFER, m_glHandle);
    }

    void calculateTangent(Vertex &v0, Vertex &v1, Vertex &v2)
    {
        glm::vec3 deltaPos1 = glm::vec3(v1.position.x, v1.position.y, v1.position.z) - glm::vec3(v0.position.x, v0.position.y, v0.position.z);
        glm::vec3 deltaPos2 = glm::vec3(v2.position.x, v2.position.y, v2.position.z) - glm::vec3(v0.position.x, v0.position.y, v0.position.z);
        glm::vec2 deltaUV1 = glm::vec2(v1.texcoord.x, v1.texcoord.y) - glm::vec2(v0.texcoord.x, v0.texcoord.y);
        glm::vec2 deltaUV2 = glm::vec2(v2.texcoord.x, v2.texcoord.y) - glm::vec2(v0.texcoord.x, v0.texcoord.y);
        float r = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV1.y * deltaUV2.x);
        glm::vec3 tangent = (deltaPos1 * deltaUV2.y   - deltaPos2 * deltaUV1.y)*r;
        glm::vec3 bitangent = (deltaPos2 * deltaUV1.x   - deltaPos1 * deltaUV2.x)*r;

        v0.tangent.x = tangent.x; v0.tangent.y = tangent.y; v0.tangent.z = tangent.z;
        v0.bitangent.x = bitangent.x; v0.bitangent.y = bitangent.y; v0.bitangent.z = bitangent.z;

        v1.tangent.x = tangent.x; v1.tangent.y = tangent.y; v1.tangent.z = tangent.z;
        v1.bitangent.x = bitangent.x; v1.bitangent.y = bitangent.y; v1.bitangent.z = bitangent.z;

        v2.tangent.x = tangent.x; v2.tangent.y = tangent.y; v2.tangent.z = tangent.z;
        v2.bitangent.x = bitangent.x; v2.bitangent.y = bitangent.y; v2.bitangent.z = bitangent.z;
    }
private:
    VertexList m_vertices;
    SurfaceList m_surfaces;
};

//------------------------------------------------------------------------------

class ScreenQuad : public Mesh {
public:
    ScreenQuad()
        : Mesh(quadVertices(), quadSurfaces())
    {}

private:
    static VertexList quadVertices() {
        VertexList vertices;
        Vertex v;
        v.position = glm::vec3(-1, -1, 0);
        v.texcoord = glm::vec2(0, 0);
        vertices.push_back(v);
        v.position = glm::vec3(-1, 1, 0);
        v.texcoord = glm::vec2(0, 1);
        vertices.push_back(v);
        v.position = glm::vec3(1, 1, 0);
        v.texcoord = glm::vec2(1, 1);
        vertices.push_back(v);
        v.position = glm::vec3(-1, -1, 0);
        v.texcoord = glm::vec2(0, 0);
        vertices.push_back(v);
        v.position = glm::vec3(1, 1, 0);
        v.texcoord = glm::vec2(1, 1);
        vertices.push_back(v);
        v.position = glm::vec3(1, -1, 0);
        v.texcoord = glm::vec2(1, 0);
        vertices.push_back(v);
        return vertices;
    }

    static SurfaceList quadSurfaces() {
        SurfaceList surfaces;
        Surface s;
        s.start = 0;
        s.count = 6;
        surfaces.push_back(s);
        return surfaces;
    }
};

//------------------------------------------------------------------------------

class Node {
public:
    Node()
        : m_translation(glm::vec3(0.0)),
          m_orientation(glm::quat()),
          m_scale(glm::vec3(1.0)),
          m_matrix(glm::mat4(1.0)),
          m_modify(true)
    {}

    void translate(const glm::vec3 &position) {
        m_translation += position;
        m_modify = true;
    }

    void setPosition(const glm::vec3 &position) {
        m_translation = position;
        m_modify = true;
    }

    void setEuler(const glm::vec3 &rotation) {
        m_orientation = glm::quat(glm::radians(rotation));
        m_modify = true;
    }

    void setQuaternion(const glm::quat &quaternion) {
        m_orientation = quaternion;
        m_modify = true;
    }

    void setScale(const glm::vec3 &scale) {
        m_scale = scale;
        m_modify = true;
    }

    void pitch(float v, bool global) {
        rotate(glm::toQuat(glm::yawPitchRoll(0.0f, glm::radians(v), 0.0f)), global);
    }

    void yaw(float v, bool global) {
        rotate(glm::toQuat(glm::yawPitchRoll(glm::radians(v), 0.0f, 0.0f)), global);
    }

    void roll(float v, bool global) {
        rotate(glm::toQuat(glm::yawPitchRoll(0.0f, 0.0f, glm::radians(v))), global);
    }

    void rotate(const glm::quat &orientation, bool global) {
        if (!global)
            m_orientation *= orientation;
        else
            m_orientation = orientation * m_orientation;
        m_modify = true;
    }

    void addScale(const glm::vec3 &scale) {
        m_scale *= scale;
        m_modify = true;
    }

    glm::vec3 position() const {
        return glm::vec3(modelMatrix()[3]);
    }

    glm::vec3 direction() const {
        return glm::vec3(modelMatrix()[2]);
    }

    glm::vec3 euler() const {
        return glm::degrees(glm::eulerAngles(glm::quat(modelMatrix())));
    }

    glm::vec3 scale() const {
        return getScaleFromMatrix(modelMatrix());
    }

    const glm::mat4 &modelMatrix() const {
        if (m_modify)
        {
            m_matrix = glm::translate(m_translation)
                    * glm::toMat4(m_orientation)
                    * glm::scale(m_scale);
            m_modify = false;
        }

        return m_matrix;
    }

    void setModelMatrix(const glm::mat4 &matrix) {
        m_modify = false;
        auto p = glm::vec3(matrix[3]);
        auto r = glm::eulerAngles(glm::quat(matrix));
        r *= 180.0f / 3.1415f;
        // TODO: scale from matrix
        setPosition(p);
        setEuler(r);
        setScale(getScaleFromMatrix(matrix));
    }

    void lookTo(const glm::vec3 &position) {
        auto m = modelMatrix();
        m_orientation = glm::toQuat(objLookAt(position, glm::vec3(m[3]), glm::vec3(m[2])));
    }

    glm::vec3 getScaleFromMatrix(const glm::mat4 &mat) const {
        glm::vec3 col1(mat[0][0], mat[0][1], mat[0][2]);
        glm::vec3 col2(mat[1][0], mat[1][1], mat[1][2]);
        glm::vec3 col3(mat[2][0], mat[2][1], mat[2][2]);

        glm::vec3 scaling;
        scaling.x = glm::length(col1);
        scaling.y = glm::length(col2);
        scaling.z = glm::length(col3);
        return scaling;
    }
private:
    static glm::vec3 localXAxis(const glm::vec3 &direction){ return glm::cross(direction, glm::vec3(0, -1, 0)); }
    static glm::vec3 localYAxis(const glm::vec3 &direction){ return glm::cross(direction, localXAxis(direction)); }

    static glm::mat4 objLookAt(const glm::vec3 &topos, const glm::vec3 &pos, const glm::vec3 &dir){
        glm::mat4 m=glm::inverse(glm::lookAt(glm::vec3(0, 0, 0), topos-pos, localYAxis(dir)));
        return m;
    }
private:
    glm::vec3 m_translation;
    glm::quat m_orientation;
    glm::vec3 m_scale;
    mutable glm::mat4 m_matrix;
    mutable bool m_modify;
};

//------------------------------------------------------------------------------

class Camera : public Node {
public:
    glm::mat4 viewMatrix() const {
        auto m = modelMatrix();
        auto p = glm::vec3(m[3]);
        auto d = glm::vec3(m[2]);
        return glm::lookAt(p, p + d, glm::vec3(0, 1, 0));
    }
};

//------------------------------------------------------------------------------

class Light : public Node {
public:
    Light()
        : m_attenuation(glm::vec3(1, 0.1, 0.01))
    {}

    void setAttenuation(const glm::vec3 &attenuation) {
        m_attenuation = attenuation;
    }

    const glm::vec3 &attenuation() const {
        return m_attenuation;
    }
private:
    glm::vec3 m_attenuation;
};

typedef std::list<Light*> LightList;

//------------------------------------------------------------------------------

class Model : public Node {
public:
    explicit Model(Mesh *mesh)
        : m_mesh(mesh)
    {}

    void render(Shader *shader) {
        if (!m_mesh)
            return;
        m_mesh->render(shader);
    }

private:
    Mesh *m_mesh;
};

typedef std::list<Model*> ModelList;

//------------------------------------------------------------------------------

class Scene {
public:
    Scene()
        : m_geometryBufferShader(new GeometryBufferShader),
          m_geometryBufferFramebuffer(new GeometryBufferFramebuffer),
          m_lightBufferShader(new LightBufferShader),
          m_screenQuad(new ScreenQuad)
    {}

    ~Scene() {

    }

    Camera *camera() const {
        return m_camera;
    }

    Model *addModel(const VertexList &vertices, const SurfaceList &surfaces) {
        auto mesh = createMesh(vertices, surfaces);
        auto model = new Model(mesh);
        m_models.push_back(model);
        return model;
    }

    Light *addLight() {
        auto light = new Light;
        m_lights.push_back(light);
        return light;
    }

    Node *addNode() {
        return new Node;
    }

    Mesh *createMesh(const VertexList &vertices, const SurfaceList &surfaces) {
        return new Mesh(vertices, surfaces);
    }

    void render() {
        clearLightBuffer();
        prepareGeometryBuffer();
        for (auto l : m_lights)
            renderLightPass(l);
    }

    glm::mat4 projectionMatrix() const {
        auto vp = viewport();
        return glm::perspective(glm::radians(60.0f), vp[2] / vp[3], 1.0f, 100.0f);
    }

private:
    void prepareGeometryBuffer() {
        glEnable(GL_DEPTH_TEST);
        m_geometryBufferFramebuffer->bind();
        glClearColor(0.5, 0.7, 0.9, 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        for (auto m : m_models) {
            m_geometryBufferShader->setupMatrices(projectionMatrix(), camera()->viewMatrix(), m->modelMatrix());
            m->render(m_geometryBufferShader);
        }
        m_geometryBufferFramebuffer->unbind();
    }

    void clearLightBuffer() {
        glClearColor(0, 0, 0, 0);
        glClear(GL_COLOR_BUFFER_BIT);
    }

    void renderLightPass(Light *light) {
        glDisable(GL_DEPTH_TEST);
        glEnable(GL_BLEND);
        glBlendFunc(1, 1);
        m_lightBufferShader->setupProjViewMatrix(projectionMatrix() * camera()->viewMatrix());
        m_lightBufferShader->setupCameraPosition(camera()->position());
        m_lightBufferShader->setupLightParams(light->position(), light->attenuation());
        m_lightBufferShader->setupGeometryBuffer(
                    m_geometryBufferFramebuffer->texture(GeometryBufferFramebuffer::MaterialDepthTexture),
                    m_geometryBufferFramebuffer->texture(GeometryBufferFramebuffer::NormalTexture),
                    m_geometryBufferFramebuffer->texture(GeometryBufferFramebuffer::ColorTexture));
        m_screenQuad->render(m_lightBufferShader);
        glDisable(GL_BLEND);
    }

private:
    Camera *m_camera = new Camera;
    ModelList m_models;
    LightList m_lights;
    GeometryBufferShader *m_geometryBufferShader;
    GeometryBufferFramebuffer *m_geometryBufferFramebuffer;
    LightBufferShader *m_lightBufferShader;
    ScreenQuad *m_screenQuad;
};

//------------------------------------------------------------------------------

Scene *scene;
Texture *textureD;
Texture *textureN;
Model *cube;

Model *addCubeModel() {

    Vertex faces[36]
    {
        Vertex(glm::vec3(-0.5, -0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, 0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(0, 1)),
        Vertex(glm::vec3(0.5, 0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, -0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, 0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, -0.5, -0.5), glm::vec3(0, 0, -1), glm::vec2(1, 0)),

        Vertex(glm::vec3(0.5, -0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, 0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(0, 1)),
        Vertex(glm::vec3(-0.5, 0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, -0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, 0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, -0.5, 0.5), glm::vec3(0, 0, 1), glm::vec2(1, 0)),

        Vertex(glm::vec3(-0.5, -0.5, 0.5), glm::vec3(-1, 0, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, 0.5, 0.5), glm::vec3(-1, 0, 0), glm::vec2(0, 1)),
        Vertex(glm::vec3(-0.5, 0.5, -0.5), glm::vec3(-1, 0, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, -0.5, 0.5), glm::vec3(-1, 0, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, 0.5, -0.5), glm::vec3(-1, 0, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, -0.5, -0.5), glm::vec3(-1, 0, 0), glm::vec2(1, 0)),

        Vertex(glm::vec3(0.5, -0.5, -0.5), glm::vec3(1, 0, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, 0.5, -0.5), glm::vec3(1, 0, 0), glm::vec2(0, 1)),
        Vertex(glm::vec3(0.5, 0.5, 0.5), glm::vec3(1, 0, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, -0.5, -0.5), glm::vec3(1, 0, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, 0.5, 0.5), glm::vec3(1, 0, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, -0.5, 0.5), glm::vec3(1, 0, 0), glm::vec2(1, 0)),

        Vertex(glm::vec3(-0.5, 0.5, -0.5), glm::vec3(0, 1, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, 0.5, 0.5), glm::vec3(0, 1, 0), glm::vec2(0, 1)),
        Vertex(glm::vec3(0.5, 0.5, 0.5), glm::vec3(0, 1, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, 0.5, -0.5), glm::vec3(0, 1, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, 0.5, 0.5), glm::vec3(0, 1, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, 0.5, -0.5), glm::vec3(0, 1, 0), glm::vec2(1, 0)),

        Vertex(glm::vec3(-0.5, -0.5, 0.5), glm::vec3(0, -1, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(-0.5, -0.5, -0.5), glm::vec3(0, -1, 0), glm::vec2(0, 1)),
        Vertex(glm::vec3(0.5, -0.5, -0.5), glm::vec3(0, -1, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(-0.5, -0.5, 0.5), glm::vec3(0, -1, 0), glm::vec2(0, 0)),
        Vertex(glm::vec3(0.5, -0.5, -0.5), glm::vec3(0, -1, 0), glm::vec2(1, 1)),
        Vertex(glm::vec3(0.5, -0.5, 0.5), glm::vec3(0, -1, 0), glm::vec2(1, 0)),
    };
    VertexList vertices(faces, faces + 36);
    SurfaceList surfaces;

    Surface s;
    s.start = 0;
    s.count = 36;
    s.material.diffuseTexture = textureD;
    s.material.normalTexture = textureN;
    surfaces.push_back(s);

    return scene->addModel(vertices, surfaces);
}

void init()
{
    textureD = new ImageTexture("stone_floor_d.jpg");
    textureN = new ImageTexture("stone_floor_n.jpg");
    scene = new Scene;
    scene->camera()->setModelMatrix(glm::translate(glm::vec3(0, 0, -2)));
    scene->addLight()->translate(glm::vec3(0, 0, -2));
    cube = addCubeModel();
}

void render()
{
    cube->yaw(1, true);
    cube->pitch(0.5, false);

    scene->render();

    glutSwapBuffers();
}

int main()
{
    int argc = 0;
    glutInit(&argc, 0);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(800, 480);
    glutCreateWindow("render");

    glewInit();
    init();
    glutIdleFunc(render);

    glutMainLoop();
    return 0;
}

